class AlreadyInjectedException(Exception):
    def __init__(self):
        super().__init__("Already injected")


class ProviderDoesNotConfiguredException(Exception):
    def __init__(self):
        super().__init__("Provider does not configured")


class DoesNotBoundException(Exception):
    def __init__(self, cls_name):
        super().__init__(f"{cls_name} is not bound")



from pytest import raises

from pythondi import (
    configure,
    configure_after_clear,
    Provider,
    Container,
)
from pythondi.exceptions import AlreadyInjectedException


class Repo:
    def __init__(self):
        pass


class SQLRepo:
    def __init__(self):
        pass


def test_configure_lazy_is_true():
    Container.clear()
    provider = Provider()
    provider.bind(Repo, SQLRepo, lazy=True)
    configure(provider=provider)
    assert isinstance(Container.get(), Provider)
    assert Container.get().bindings[Repo] == SQLRepo
    with raises(AlreadyInjectedException):
        configure(provider=provider)
    Container.clear()


def test_configure_lazy_is_false():
    Container.clear()
    provider = Provider()
    provider.bind(Repo, SQLRepo, lazy=False)
    configure(provider=provider)
    assert isinstance(Container.get(), Provider)
    assert Container.get().bindings[Repo] != SQLRepo
    with raises(AlreadyInjectedException):
        configure(provider=provider)
    Container.clear()


def test_configure_after_clear_lazy_is_true():
    provider = Provider()
    provider.bind(Repo, SQLRepo, lazy=True)
    configure_after_clear(provider=provider)

    provider = Provider()
    provider.bind(SQLRepo, Repo, lazy=True)
    configure_after_clear(provider=provider)
    with raises(KeyError):
        assert isinstance(Container.get(), Provider)
        assert Container.get().bindings[Repo] == SQLRepo


def test_configure_after_clear_lazy_is_false():
    provider = Provider()
    provider.bind(Repo, SQLRepo, lazy=False)
    configure_after_clear(provider=provider)

    provider = Provider()
    provider.bind(SQLRepo, Repo, lazy=False)
    configure_after_clear(provider=provider)
    with raises(KeyError):
        assert isinstance(Container.get(), Provider)
        assert Container.get().bindings[Repo] != SQLRepo



import inspect
import threading
from functools import wraps
from typing import Optional, NoReturn

from pythondi.container import Container
from pythondi.exceptions import (
    ProviderDoesNotConfiguredException,
    AlreadyInjectedException,
)
from pythondi.provider import Provider

_LOCK = threading.RLock()


def configure(provider: Provider) -> Optional[NoReturn]:
    """Configure provider to container"""
    with _LOCK:
        if Container.get():
            raise AlreadyInjectedException

        Container.set(provider=provider)


def configure_after_clear(provider: Provider) -> None:
    """Clear existing provider and configure new provider"""
    with _LOCK:
        if Container.get():
            clear()

        Container.set(provider=provider)


def clear() -> None:
    """Clear current provider"""
    _container = Container

    with _LOCK:
        _container.clear()


def inject(**params):
    def inner_func(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            _provider = Container.get()
            if not _provider:
                raise ProviderDoesNotConfiguredException

            # Case of auto injection
            if not params:
                annotations = inspect.getfullargspec(func).annotations
                for k, v in annotations.items():
                    if v in _provider.bindings and k not in kwargs:
                        replacement = _provider.bindings[v]
                        if inspect.isclass(replacement):
                            replacement = replacement()

                        kwargs[k] = replacement
            # Case of manual injection
            else:
                for k, v in params.items():
                    replacement = v
                    if inspect.isclass(replacement):
                        replacement = replacement()

                    kwargs[k] = replacement

            if inspect.iscoroutinefunction(func):
                async def _inject(*args, **kwargs):
                    return await func(*args, **kwargs)
            else:
                def _inject(*args, **kwargs):
                    return func(*args, **kwargs)

            return _inject(*args, **kwargs)
        return wrapper
    return inner_func


__all__ = [
    "Provider",
    "configure",
    "configure_after_clear",
    "inject",
]



import inspect

import pytest

from pythondi import inject, Provider, configure_after_clear


class Repo:
    def __init__(self):
        pass


class SQLRepo:
    def __init__(self):
        pass


class Usecase:
    def __init__(self):
        pass


class UserUsecase:
    def __init__(self):
        pass


def test_sync_inject_without_parameter_lazy_is_true():
    provider = Provider()
    provider.bind(Repo, SQLRepo, lazy=True)
    configure_after_clear(provider)

    @inject()
    def func(repo: Repo):
        assert isinstance(repo, SQLRepo)

    func()


def test_sync_inject_without_parameter_lazy_is_true_singleton():
    provider = Provider()
    provider.bind(Repo, SQLRepo, lazy=True)
    configure_after_clear(provider)

    @inject()
    def func(repo: Repo):
        return repo

    @inject()
    def func2(repo: Repo):
        return repo

    repo1 = func()
    repo2 = func2()

    assert repo1 != repo2


def test_sync_inject_without_parameter_lazy_is_false():
    provider = Provider()
    provider.bind(Repo, SQLRepo, lazy=False)
    configure_after_clear(provider)

    @inject()
    def func(repo: Repo):
        assert isinstance(repo, SQLRepo)

    func()


def test_sync_inject_without_parameter_lazy_is_false_singleton():
    provider = Provider()
    provider.bind(Repo, SQLRepo, lazy=False)
    configure_after_clear(provider)

    @inject()
    def func(repo: Repo):
        return repo

    @inject()
    def func2(repo: Repo):
        return repo

    repo1 = func()
    repo2 = func2()

    assert repo1 == repo2
    assert not inspect.isclass(repo1)
    assert not inspect.isclass(repo2)


def test_sync_inject_without_parameter_multiple_bind_lazy_true():
    provider = Provider()
    provider.bind(Repo, SQLRepo, lazy=True)
    provider.bind(Usecase, UserUsecase, lazy=True)
    configure_after_clear(provider)

    @inject()
    def func(repo: Repo, usecase: Usecase):
        assert isinstance(repo, SQLRepo)
        assert isinstance(usecase, UserUsecase)

    func()


def test_sync_inject_without_parameter_multiple_bind_lazy_false():
    provider = Provider()
    provider.bind(Repo, SQLRepo, lazy=False)
    provider.bind(Usecase, UserUsecase, lazy=False)
    configure_after_clear(provider)

    @inject()
    def func(repo: Repo, usecase: Usecase):
        assert isinstance(repo, SQLRepo)
        assert isinstance(usecase, UserUsecase)

    func()


def test_sync_inject_with_parameter():
    provider = Provider()
    provider.bind(Repo, SQLRepo)
    configure_after_clear(provider)

    @inject(repo=SQLRepo)
    def func(repo):
        assert isinstance(repo, SQLRepo)

    func()


def test_sync_inject_with_parameter_multiple_bind():
    provider = Provider()
    configure_after_clear(provider)

    @inject(repo=SQLRepo, usecase=UserUsecase)
    def func(repo, usecase):
        assert isinstance(repo, SQLRepo)
        assert isinstance(usecase, UserUsecase)

    func()


@pytest.mark.asyncio
async def test_async_inject_without_parameter_lazy_true():
    provider = Provider()
    provider.bind(Repo, SQLRepo, lazy=True)
    configure_after_clear(provider)

    @inject()
    async def func(repo: Repo):
        assert isinstance(repo, SQLRepo)

    await func()


@pytest.mark.asyncio
async def test_async_inject_without_parameter_lazy_false():
    provider = Provider()
    provider.bind(Repo, SQLRepo, lazy=False)
    configure_after_clear(provider)

    @inject()
    async def func(repo: Repo):
        assert isinstance(repo, SQLRepo)

    await func()


@pytest.mark.asyncio
async def test_async_inject_without_parameter_multiple_bind_lazy_true():
    provider = Provider()
    provider.bind(Repo, SQLRepo, lazy=True)
    provider.bind(Usecase, UserUsecase, lazy=True)
    configure_after_clear(provider)

    @inject()
    async def func(repo: Repo, usecase: Usecase):
        assert isinstance(repo, SQLRepo)
        assert isinstance(usecase, UserUsecase)

    await func()


@pytest.mark.asyncio
async def test_async_inject_without_parameter_multiple_bind_lazy_false():
    provider = Provider()
    provider.bind(Repo, SQLRepo, lazy=False)
    provider.bind(Usecase, UserUsecase, lazy=False)
    configure_after_clear(provider)

    @inject()
    async def func(repo: Repo, usecase: Usecase):
        assert isinstance(repo, SQLRepo)
        assert isinstance(usecase, UserUsecase)

    await func()


@pytest.mark.asyncio
async def test_async_inject_with_parameter():
    provider = Provider()
    provider.bind(Repo, SQLRepo)
    configure_after_clear(provider)

    @inject(repo=SQLRepo)
    async def func(repo):
        assert isinstance(repo, SQLRepo)

    await func()


@pytest.mark.asyncio
async def test_async_inject_with_parameter_multiple_bind():
    provider = Provider()
    configure_after_clear(provider)

    @inject(repo=SQLRepo, usecase=UserUsecase)
    async def func(repo, usecase):
        assert isinstance(repo, SQLRepo)
        assert isinstance(usecase, UserUsecase)

    await func()



from typing import Optional

from pythondi.provider import Provider


class Container:
    _provider = None

    @classmethod
    def set(cls, provider: Provider) -> None:
        """Set provider"""
        cls._provider = provider

    @classmethod
    def get(cls) -> Optional[Provider]:
        """Get current provider"""
        return cls._provider

    @classmethod
    def clear(cls) -> None:
        """Clear provider"""
        cls._provider = None



import inspect

from pytest import raises

from pythondi import Provider
from pythondi.exceptions import DoesNotBoundException


class Repo:
    def __init__(self):
        pass


class SQLRepo:
    def __init__(self):
        pass


def test_bind_lazy_is_true():
    provider = Provider()
    provider.bind(Repo, SQLRepo, lazy=True)
    assert provider.bindings[Repo] == SQLRepo
    assert inspect.isclass(provider.bindings[Repo])


def test_bind_lazy_is_false():
    provider = Provider()
    provider.bind(Repo, SQLRepo, lazy=False)
    assert provider.bindings[Repo] != SQLRepo
    assert not inspect.isclass(provider.bindings[Repo])


def test_unbind():
    provider = Provider()
    with raises(DoesNotBoundException):
        provider.unbind(interface=int)

    provider.bind(int, str)
    provider.unbind(interface=int)
    assert provider.bindings == {}


def test_clear_bindings():
    provider = Provider()
    provider.bind(int, str)
    provider.bind(list, dict)
    provider.clear_bindings()
    assert provider.bindings == {}



from typing import Optional, NoReturn

from pythondi.exceptions import DoesNotBoundException


class Provider:
    def __init__(self):
        self._bindings = {}

    def bind(self, interface=None, impl=None, lazy: bool = False) -> Optional[NoReturn]:
        if lazy is False:
            impl = impl()

        self._bindings[interface] = impl

    def unbind(self, interface) -> Optional[NoReturn]:
        """Unbind class"""
        try:
            self._bindings.pop(interface)
        except KeyError:
            raise DoesNotBoundException(cls_name=interface.__class__.__name__)

    def clear_bindings(self) -> None:
        """Clear bindings"""
        self._bindings = {}

    @property
    def bindings(self) -> dict:
        """Get current bindings"""
        return self._bindings
